# -*- coding: utf-8 -*-
"""
Rhizom - Relationship grapher

Copyright (C) 2015  Aurelien Bompard <aurelien@bompard.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import unicode_literals, print_function

import browserid
from flask import request, g, render_template, abort, redirect, url_for, flash, jsonify
from flask.ext.login import login_user, logout_user, current_user
from sqlalchemy import or_
from sqlalchemy.orm.exc import NoResultFound
from . import app
from .models import Graph, Person, Relationship, User, Permission, PermissionLevel, RelationshipType
from .lib import compute_node_circles, compute_node_branch, require_permission_level
from .forms import (NewRelationship, NewAccess, NewGraph, CopyGraph,
                    RelationshipTypes, GraphProperties, UserForm)


@app.route('/login', methods=["GET", "POST"])
def login():
    if request.method == "POST":
        # La requête doit avoir une assertion à vérifier
        if 'assertion' not in request.form:
            abort(400)
        data = browserid.verify(request.form['assertion'],
                                app.config["BROWSERID_AUDIENCE"])
        user = g.db.query(User).get(data['email'])
        if user is None:
            if g.db.query(User).count() == 0:
                user = User(email=data["email"], name="Admin")
                g.db.add(user)
                g.db.commit()
            else:
                flash("Echec de la connexion, utilisateur inconnu.", "warning")
                return redirect(url_for("index"))
        login_user(user)
        flash("Bonjour {} ! :-)".format(user.name), "success")
        # no redirect here, the JS will follow it in the AJAX request and load
        # the page twice.
        return request.args.get("next") or url_for("index")
    if current_user.is_authenticated():
        print("Already authenticated")
    return redirect(request.args.get("next") or url_for("index"))


@app.route('/logout', methods=["GET", "POST"])
def logout():
    if current_user.is_authenticated():
        name = current_user.name
        logout_user()
        flash("Déconnexion réussie, à bientôt {} !".format(name), "success")
    # no redirect here, the JS will follow it in the AJAX request and load the
    # page twice.
    return url_for('index')


@app.route('/', methods=["GET", "POST"])
def index():
    if not current_user.is_authenticated():
        graphs = []
    else:
        if current_user.is_master:
            graphs = g.db.query(Graph).order_by(Graph.id).all()
        else:
            graphs = g.db.query(Graph).join(Permission).filter(
                Permission.user_email == current_user.email,
                Permission.level >= PermissionLevel.view
                ).order_by(Graph.id).all()
    new_form = NewGraph(prefix="newgraph")
    copy_form = CopyGraph(prefix="copygraph")
    copy_form.existing.choices = [ (str(graph.id), graph.name) for graph in graphs ]
    if request.method == "POST":
        if request.form.get("formname") == "new" and new_form.validate():
            graph = Graph(name=new_form.name.data)
            g.db.add(graph)
            g.db.flush() # otherwise graph.id is None
            g.db.add(Permission(graph_id=graph.id,
                                user_email=current_user.email,
                                level=PermissionLevel.admin))
            g.db.commit()
            flash("Graphe créé", "success")
            return redirect(url_for("admin", graph_id=graph.id))
        if request.form.get("formname") == "copy" and copy_form.validate():
            existing = g.db.query(Graph).get(copy_form.existing.data)
            graph = Graph(name="Copie de {}".format(existing.name))
            g.db.add(graph)
            g.db.flush() # otherwise graph.id is None
            graph.copy_from(existing)
            # The copying user is the admin of the new graph
            myperm = g.db.query(Permission).get((graph.id, current_user.email))
            myperm.level = PermissionLevel.admin
            g.db.commit()
            flash("Graphe copié", "success")
            return redirect(url_for("admin", graph_id=graph.id))
    return render_template("index.html",
        graphs=graphs, new_form=new_form, copy_form=copy_form)


@app.route('/graph/<int:graph_id>')
def graph(graph_id):
    graph = g.db.query(Graph).get(graph_id)
    if not graph:
        return render_template("error.html",
                               code=404, message="Graphe introuvable"), 404
    if not require_permission_level(graph_id, PermissionLevel.view):
        return render_template("error.html",
                               code=403, message="Accès interdit"), 403
    rel_types = g.db.query(RelationshipType).filter_by(
        graph_id=graph_id).order_by(RelationshipType.name).all()
    return render_template("view.html", graph=graph, rel_types=rel_types)


@app.route('/api/graph/<int:graph_id>/data.js')
def data(graph_id):
    graph = g.db.query(Graph).get(graph_id)
    if not graph:
        abort(404)
    if not require_permission_level(graph_id, PermissionLevel.view):
        abort(403)
    data = {"nodes": [], "links": [], "center": None,
            "anonymous": graph.anonymous}
    node_index = {}
    for index, person in enumerate(g.db.query(Person).filter_by(
        graph_id=graph_id).order_by(Person.id)):
        node = {"id": person.id}
        #node["fixed"] = True
        if person.center:
            node.update({"size": 15, "fixed": True, "center": True})
            data["center"] = index
        node["name"] = person.name
        data["nodes"].append(node)
        node_index[person.id] = index
    siblings = {}
    for rel in g.db.query(Relationship).filter_by(graph_id=graph_id).all():
        link = {"source": node_index[rel.source_id],
                "target": node_index[rel.target_id],
                "css": rel.type.cssname.lower(),
                "dotted": rel.dotted}
        siblings_count = g.db.query(Relationship).filter_by(
            source_id=rel.source_id, target_id=rel.target_id
            ).count()
        link["siblings"] = siblings_count
        if siblings_count > 0:
            if (rel.source_id, rel.target_id) in siblings:
                siblings[(rel.source_id, rel.target_id)] += 1
                sibling_id = siblings[(rel.source_id, rel.target_id)]
            elif (rel.target_id, rel.source_id) in siblings:
                siblings[(rel.target_id, rel.source_id)] += 1
                sibling_id = siblings[(rel.target_id, rel.source_id)]
            else:
                siblings[(rel.source_id, rel.target_id)] = sibling_id = 0
            link["sibling_id"] = sibling_id
        data["links"].append(link)
    if data["center"] is not None:
        compute_node_circles(data["nodes"], data["links"], data["center"])
        compute_node_branch(data["nodes"], data["links"])
    return jsonify(**data)


@app.route('/graph/<int:graph_id>/edit', methods=["GET", "POST"])
def edit(graph_id):
    graph = g.db.query(Graph).get(graph_id)
    if not graph:
        abort(404)
    if not require_permission_level(graph_id, PermissionLevel.edit):
        return render_template("error.html",
                               code=403, message="Accès interdit"), 403
    newrel_form = NewRelationship()
    rel_types = [ rt.name for rt in g.db.query(RelationshipType.name).filter_by(
        graph_id=graph_id).order_by(RelationshipType.name) ]
    newrel_form.rtype.choices = [ (rt, rt) for rt in rel_types ]
    context = dict(
        graph = graph,
        newrel_form = newrel_form,
        relationships = g.db.query(Relationship).filter_by(
            graph_id=graph_id).order_by(Relationship.source_id).all(),
        rel_types = rel_types,
    )
    if request.method == "POST":
        if request.form.get("action") == "delete":
            sid = int(request.form["source"])
            tid = int(request.form["target"])
            ltype = request.form["origltype"]
            link = g.db.query(Relationship).filter_by(
                source_id=sid, target_id=tid, type_name=ltype).one()
            g.db.delete(link)
            g.db.flush() # flushing is required for the next step
            # Cleanup orphans
            for pid in (sid, tid):
                if g.db.query(Relationship).filter(or_(
                    Relationship.source_id == pid,
                    Relationship.target_id == pid)).count() == 0:
                    g.db.delete(g.db.query(Person).get(pid))
            g.db.commit()
            #flash("Relation supprimée", "success")
            return jsonify({"status": "OK"})
            #return redirect(url_for("edit", graph_id=graph_id))
        elif request.form.get("action") == "edit":
            sid = int(request.form["source"])
            tid = int(request.form["target"])
            ltype = request.form["origltype"]
            newltype = request.form["ltype"]
            dotted = request.form["dotted"] not in ("", "false")
            if ltype != newltype and g.db.query(Relationship).filter_by(
                source_id=sid, target_id=tid, type_name=newltype).count():
                return jsonify({"status": "error",
                                "message": "Cette relation existe déjà"})
            link = g.db.query(Relationship).filter_by(
                source_id=sid, target_id=tid, type_name=ltype).one()
            link.type_name = newltype
            link.dotted = dotted
            g.db.commit()
            flash("Relation modifiée", "success")
            return jsonify({"status": "OK"})
        elif request.form.get("action") == "add":
            if not newrel_form.validate():
                return render_template("edit.html", **context)
            try:
                source = g.db.query(Person).filter_by(
                    graph_id=graph_id, name=newrel_form.source.data).one()
            except NoResultFound:
                source = Person(graph_id=graph_id, name=newrel_form.source.data)
                g.db.add(source)
            try:
                target = g.db.query(Person).filter_by(
                    graph_id=graph_id, name=newrel_form.target.data).one()
            except NoResultFound:
                target = Person(graph_id=graph_id, name=newrel_form.target.data)
                g.db.add(target)
            g.db.flush() # flushing is required for the next step
            if source.id > target.id:
                source, target = target, source # the source is always the lowest id
            if g.db.query(Relationship).get(
                (source.id, target.id, newrel_form.rtype.data)) is not None:
                flash("Cette relation existe déjà", "warning")
                return render_template("edit.html", **context)
            newlink = Relationship(
                source_id=source.id, target_id=target.id,
                type_name=newrel_form.rtype.data, graph_id=graph.id,
                dotted=newrel_form.dotted.data)
            g.db.add(newlink)
            g.db.commit()
            flash("Relation ajoutée", "success")
            return redirect(url_for("edit", graph_id=graph_id))
    return render_template("edit.html", **context)


@app.route('/graph/<int:graph_id>/admin', methods=["GET", "POST"])
def admin(graph_id):
    graph = g.db.query(Graph).get(graph_id)
    if not graph:
        abort(404)
    if not require_permission_level(graph_id, PermissionLevel.admin):
        return render_template("error.html",
                               code=403, message="Accès interdit"), 403
    newaccess_form = NewAccess(prefix="newaccess")
    graphprops_form = GraphProperties(obj=graph, prefix="graphprops")
    graphprops_form.center_id.choices = [("0", "(personne)")] + [
        (str(p.id), p.name) for p in g.db.query(Person).filter_by(graph_id=graph_id
        ).order_by(Person.name) ]
    reltypes_form = RelationshipTypes(prefix="reltypes")
    perm_choices = [ (int(getattr(PermissionLevel, l)), n) for l, n in
                  ( ("view", "Voir"), ("edit", "Modifier"),
                    ("admin", "Administrer") ) ]
    newaccess_form.level.choices = perm_choices
    context = dict(
        graph = graph,
        graphprops_form = graphprops_form,
        reltypes_form = reltypes_form,
        newaccess_form = newaccess_form,
        permissions = g.db.query(Permission).join(User).filter(
            Permission.graph_id == graph_id).order_by(User.name).all(),
        rel_types = g.db.query(RelationshipType).filter_by(
            graph_id=graph_id).order_by(RelationshipType.name).all(),
        perm_choices = perm_choices,
    )
    if request.method == "POST":
        # Graph: properties edition
        if request.form.get("formname") == "graphprops":
            if not graphprops_form.validate():
                return render_template("admin.html", **context)
            graph.name = graphprops_form.name.data
            graph.center_id = int(graphprops_form.center_id.data)
            g.db.commit()
            flash("Graphe mis à jour", "success")
            return redirect(url_for("admin", graph_id=graph_id))
        # Permissions
        elif request.form.get("formname") == "permissions":
            if request.form.get("action") == "add":
                if not newaccess_form.validate():
                    return render_template("admin.html", **context)
                user = g.db.query(User).get(newaccess_form.email.data)
                if user is None:
                    user = User(email=newaccess_form.email.data,
                                name=newaccess_form.name.data)
                    g.db.add(user)
                perm = Permission(graph_id=graph_id,
                                  user_email=newaccess_form.email.data,
                                  level=newaccess_form.level.data)
                g.db.add(perm)
                g.db.commit()
                flash("Accès ajouté", "success")
                return redirect(url_for("admin", graph_id=graph_id))
            elif request.form.get("action") == "delete":
                perm = g.db.query(Permission).get((graph_id, request.form["email"]))
                if perm is None:
                    return jsonify({"status": "error",
                                    "message": "Invalid permission"})
                user = perm.user
                g.db.delete(perm)
                g.db.flush() # flushing is required for the next step
                if len(user.permissions) == 0:
                    g.db.delete(user) # no permissions left
                g.db.commit()
                #flash("Accès supprimé", "success")
                return jsonify({"status": "OK"})
            elif request.form.get("action") == "edit":
                perm = g.db.query(Permission).get((graph_id, request.form["email"]))
                if perm is None:
                    return jsonify({"status": "error",
                                    "message": "Invalid relationship type"})
                perm.level = request.form["level"]
                perm.user.name = request.form["name"]
                g.db.commit()
                flash("Utilisateur modifié", "success")
                return jsonify({"status": "OK"})
        ## Graph: relationship types
        elif request.form.get("formname") == "reltypes":
            if request.form.get("action") == "add":
                if not reltypes_form.validate():
                    return render_template("admin.html", **context)
                reltype = RelationshipType(graph_id=graph_id,
                                           name=reltypes_form.name.data,
                                           color=reltypes_form.color.data)
                for existing_types in g.db.query(RelationshipType
                    ).filter_by(graph_id=graph_id):
                    if reltype.cssname == existing_types.cssname:
                        reltypes_form.name.errors.append("Ce nom est déjà utilisé")
                        return render_template("admin.html", **context)
                g.db.add(reltype)
                g.db.commit()
                flash("Type de relation ajouté", "success")
                return redirect(url_for("admin", graph_id=graph_id))
            elif request.form.get("action") == "delete":
                reltype = g.db.query(RelationshipType).get(
                    (graph_id, request.form["origname"]))
                if reltype is None:
                    return jsonify({"status": "error",
                                    "message": "Invalid relationship type"})
                g.db.delete(reltype)
                g.db.commit()
                #flash("Type de relation supprimé", "success")
                return jsonify({"status": "OK"})
            elif request.form.get("action") == "edit":
                reltype = g.db.query(RelationshipType).get(
                    (graph_id, request.form["origname"]))
                if reltype is None:
                    return jsonify({"status": "error",
                                    "message": "Invalid relationship type"})
                reltype.name = request.form["name"]
                reltype.color = request.form["color"]
                g.db.commit()
                flash("Type de relation modifié", "success")
                return jsonify({"status": "OK"})
            else:
                flash("Invalid request", "danger")
        elif request.form.get("formname") == "graphdel":
            g.db.delete(graph)
            g.db.commit()
            flash("Graphe supprimé", "success")
            return redirect(url_for("index"))
    return render_template("admin.html", **context)


@app.route('/users', methods=["GET", "POST"])
def users():
    if not current_user.is_master:
        return render_template("error.html",
                               code=403, message="Accès interdit"), 403
    user_form = UserForm()
    context = dict(
        user_form = user_form,
        users = g.db.query(User).order_by(User.name).all(),
    )
    if request.method == "POST":
        if request.form.get("action") == "delete":
            user = g.db.query(User).get(request.form["email"])
            g.db.delete(user)
            g.db.commit()
            #flash("Utilisateur supprimé", "success")
            return jsonify({"status": "OK"})
            #return redirect(url_for("users"))
        elif request.form.get("action") == "edit":
            user = g.db.query(User).get(request.form["email"])
            user.name = request.form["name"]
            g.db.commit()
            flash("Utilisateur modifié", "success")
            return jsonify({"status": "OK"})
        elif request.form.get("action") == "add" and user_form.validate():
            user = User(name=user_form.name.data, email=user_form.email.data)
            g.db.add(user)
            g.db.commit()
            flash("Utilisateur ajouté", "success")
            return redirect(url_for("users"))
    return render_template("users.html", **context)


@app.route('/errors/<int:code>')
def error(code):
    return render_template("error.html", code=code), code



# -*- coding: utf-8 -*-
"""
Rhizom - Relationship grapher

Copyright (C) 2015  Aurelien Bompard <aurelien@bompard.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import absolute_import, unicode_literals, print_function

import os
from flask import Flask, g
from flask.ext.login import LoginManager
from sqlalchemy.orm.exc import NoResultFound

app = Flask(__name__)

app.config.from_object('rhizom.defaults')
if 'RHIZOM_SETTINGS' in os.environ:
    app.config.from_envvar('RHIZOM_SETTINGS')

login_manager = LoginManager()
login_manager.init_app(app)
login_manager.login_view = "index"
login_manager.login_message = "Merci de vous connecter pour accéder à cette page."
login_manager.login_message_category = "info"

from .models import User

@login_manager.user_loader
def load_user(userid):
    return User.query.get(userid)


#
# Database
#

@app.before_request
def before_request():
    from .database import get_session
    g.db = get_session(app.config["SQLALCHEMY_DATABASE_URI"])

@app.teardown_appcontext
def shutdown_session(exception=None):
    if hasattr(g, "db"):
        g.db.remove()

from sqlalchemy.engine import Engine
from sqlalchemy import event as sa_event
from sqlite3 import Connection as SQLite3Connection
@sa_event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    if isinstance(dbapi_connection, SQLite3Connection):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()


# Views
from .views import *


if __name__ == '__main__':
    app.run()



# -*- coding: utf-8 -*-
"""
Rhizom - Relationship grapher

Copyright (C) 2015  Aurelien Bompard <aurelien@bompard.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import unicode_literals, print_function

import re

from enum import IntEnum

from sqlalchemy import Column, Integer, Unicode, Boolean, ForeignKey, or_
from sqlalchemy.orm import relationship, object_session
from sqlalchemy.schema import ForeignKeyConstraint
from sqlalchemy.sql.expression import false
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm.exc import NoResultFound

from flask.ext.login import UserMixin
from . import app


Base = declarative_base()


class DumpableMixin:

    dump_attr = ()

    def as_dict(self):
        return dict((a, getattr(self, a)) for a in self.dump_attr)


class Graph(Base):

    __tablename__ = 'graphs'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(254), index=True)
    anonymous = Column(Boolean, server_default=false(), nullable=False)
    persons = relationship("Person",
        backref="graph", cascade="all, delete-orphan", passive_deletes=True)
    permissions = relationship("Permission", backref="graph")

    @property
    def center_id(self):
        result = object_session(self).query(Person.id).filter_by(
            graph_id=self.id, center=True).first()
        if result is None:
            return None
        return result.id
    @center_id.setter
    def center_id(self, person_id):
        if not person_id: # 0 or None
            for p in object_session(self).query(Person).filter_by(
                graph_id=self.id, center=True):
                p.center = False
        else:
            person = object_session(self).query(Person).get(person_id)
            if person.graph_id != self.id:
                raise ValueError
            person.center = True

    def copy_from(self, graph):
        db = object_session(self)
        self.anonymous = graph.anonymous
        for perm in db.query(Permission).filter_by(graph_id=graph.id):
            db.add(Permission(graph_id=self.id,
                user_email=perm.user_email, level=perm.level))
        for person in db.query(Person).filter_by(graph_id=graph.id):
            db.add(Person(graph_id=self.id,
                name=person.name, center=person.center))
        for reltype in db.query(RelationshipType).filter_by(graph_id=graph.id):
            db.add(RelationshipType(graph_id=self.id,
                name=reltype.name, color=reltype.color))
        db.flush()
        for rel in db.query(Relationship).filter_by(graph_id=graph.id):
            source = db.query(Person).filter_by(
                graph_id=self.id, name=rel.source.name).one()
            target = db.query(Person).filter_by(
                graph_id=self.id, name=rel.target.name).one()
            db.add(Relationship(graph_id=self.id,
                source=source, target=target, type_name=rel.type_name))


class Person(Base, DumpableMixin):

    __tablename__ = 'persons'

    id = Column(Integer, primary_key=True)
    graph_id = Column(Integer,
        ForeignKey("graphs.id", ondelete="cascade", onupdate="cascade"),
        nullable=False)
    name = Column(Unicode(254), index=True)
    center = Column(Boolean, server_default=false(), nullable=False)

    dump_attr = ("id", "name", "center")

    def __repr__(self):
        return '<Person %r>' % (self.name)

    @property
    def relationships(self):
        targets = object_session(self).query(Person).join(
            Relationship, Person.id==Relationship.target_id
            ).filter(Relationship.source_id == self.id)
        sources = object_session(self).query(Person).join(
            Relationship, Person.id==Relationship.source_id
            ).filter(Relationship.target_id == self.id)
        return targets.union(sources).all()


class Relationship(Base, DumpableMixin):

    __tablename__ = "relationships"
    __table_args__ = (
        ForeignKeyConstraint(
            ['graph_id', 'type_name'],
            ['relationship_types.graph_id', 'relationship_types.name'],
            ondelete="cascade", onupdate="cascade"
        ),
    )

    source_id = Column(Integer,
        ForeignKey("persons.id", ondelete="cascade", onupdate="cascade"),
        primary_key=True)
    target_id = Column(Integer,
        ForeignKey("persons.id", ondelete="cascade", onupdate="cascade"),
        primary_key=True)
    type_name = Column(Unicode(32), primary_key=True)
    graph_id = Column(Integer, ForeignKey("graphs.id"))
    dotted = Column(Boolean, server_default=false(), nullable=False)
    type = relationship("RelationshipType")
    source = relationship("Person", foreign_keys=[source_id])
    target = relationship("Person", foreign_keys=[target_id])

    dump_attr = ("source_id", "target_id", "type", "dotted")

    def __repr__(self):
        return '<Relationship %r - %r (%r)>' % (
                self.source_id, self.target_id, self.type)


class RelationshipType(Base):

    __tablename__ = "relationship_types"

    graph_id = Column(Integer,
        ForeignKey("graphs.id", ondelete="cascade", onupdate="cascade"),
        primary_key=True, nullable=False)
    name = Column(Unicode(32), primary_key=True, nullable=False)
    color = Column(Unicode(32))

    @property
    def cssname(self):
        return re.sub('[^a-z0-9-]', '', self.name.lower())

    @property
    def rel_count(self):
        return object_session(self).query(Relationship).filter_by(
            graph_id=self.graph_id, type_name=self.name).count()


class User(Base, UserMixin, DumpableMixin):

    __tablename__ = 'users'

    email = Column(Unicode(120), primary_key=True)
    name = Column(Unicode(254))
    permissions = relationship("Permission",
        backref="user", cascade="all, delete-orphan", passive_deletes=True)

    dump_attr = ("email", "name")

    def __repr__(self):
        return '<User %r>' % (self.email)

    def get_id(self):
        return self.email

    def has_perm(self, graph, level):
        if self.is_master:
            return True
        try:
            level = object_session(self).query(Permission.level
                ).with_parent(self).filter_by(
                graph_id=graph.id).one().level
        except NoResultFound:
            return False
        return level >= PermissionLevel(level)

    @property
    def is_master(self):
        return self.email in app.config.get("ADMINS", [])

    def perm_for(self, graph):
        try:
            level = object_session(self).query(Permission.level
                ).with_parent(self).filter_by(
                graph_id=graph.id).one().level
        except NoResultFound:
            return None
        return PermissionLevel(level)


class Permission(Base):

    __tablename__ = 'permissions'

    graph_id = Column(Integer,
        ForeignKey("graphs.id", ondelete="cascade", onupdate="cascade"),
        primary_key=True)
    user_email = Column(Unicode(120),
        ForeignKey("users.email", ondelete="cascade", onupdate="cascade"),
        primary_key=True)
    level = Column(Integer)

    @property
    def level_as_string(self):
        return PermissionLevel(self.level).name


class PermissionLevel(IntEnum):
    view   = 1
    edit   = 2
    admin  = 3



# -*- coding: utf-8 -*-
"""
Rhizom - Relationship grapher

Copyright (C) 2015  Aurelien Bompard <aurelien@bompard.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import unicode_literals, print_function

import flask
from flask_wtf import Form
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from wtforms import (StringField, HiddenField, SelectField, BooleanField,
                     validators, widgets, FieldList)
from . import app
from .models import Person, User, Permission, PermissionLevel


def _addclass(kwargs, cssclass):
    if "class" in kwargs:
        kwargs["class"] = "{} {}".format(kwargs["class"], cssclass)
    else:
        kwargs["class"] = cssclass


class BootstrapTextInput(widgets.TextInput):

    def __call__(self, field, **kwargs):
        _addclass(kwargs, "form-control")
        return super(BootstrapTextInput, self).__call__(field, **kwargs)


class InlineTextInput(BootstrapTextInput):

    def __call__(self, field, **kwargs):
        kwargs["placeholder"] = field.label.text
        return super(InlineTextInput, self).__call__(field, **kwargs)


class SmallTextInput(InlineTextInput):

    def __call__(self, field, **kwargs):
        _addclass(kwargs, "input-sm")
        return super(SmallTextInput, self).__call__(field, **kwargs)


def strip(s):
    if s:
        return s.strip()
    else:
        return ""


class NewRelationship(Form):
    source = StringField("Nom", filters=[strip],
        validators=[validators.InputRequired(), validators.Length(max=254)],
        widget=SmallTextInput())
    target = StringField("Nom", filters=[strip],
        validators=[validators.InputRequired(), validators.Length(max=254)],
        widget=SmallTextInput())
    rtype =  SelectField("Type",
        validators=[validators.InputRequired()])
    dotted = BooleanField("Pointillé")


class NewAccess(Form):
    name = StringField("Nom", filters=[strip],
        validators=[validators.InputRequired(), validators.Length(max=254)],
        widget=SmallTextInput())
    email = StringField("Email de connexion", filters=[strip],
        validators=[validators.InputRequired(), validators.Length(max=120),
                    validators.Email()],
        widget=SmallTextInput(input_type="email"))
    level = SelectField("Niveau", filters=[int],
        validators=[validators.InputRequired()])


class NewGraph(Form):
    name = StringField("Nom", filters=[strip],
        validators=[validators.InputRequired(), validators.Length(max=254)],
        widget=InlineTextInput())

class CopyGraph(Form):
    existing = SelectField(filters=[strip],
        validators=[validators.InputRequired()])


class GraphProperties(Form):
    name = StringField("Nom", filters=[strip],
        validators=[validators.InputRequired(), validators.Length(max=254)],
        widget=BootstrapTextInput())
    center_id = SelectField("Centre", filters=[strip],
        validators=[validators.InputRequired()])
    anonymous = BooleanField("Anonyme par défaut")


class RelationshipTypes(Form):
    name = StringField("Nom", filters=[strip],
        validators=[validators.InputRequired(), validators.Length(max=32)],
        widget=SmallTextInput())
    color = StringField("Couleur", filters=[strip],
        validators=[validators.InputRequired(), validators.Length(max=32)],
        widget=SmallTextInput(input_type="color"))


def new_email(form, field):
    if flask.g.db.query(User).filter_by(
        email=field.data).count() != 0:
        raise validators.ValidationError("Email déjà utilisé.")

class UserForm(Form):
    name = StringField("Nom", filters=[strip],
        validators=[validators.InputRequired(), validators.Length(max=254)],
        widget=SmallTextInput())
    email = StringField("Email de connexion", filters=[strip],
        validators=[validators.InputRequired(), validators.Length(max=120),
                    validators.Email(), new_email],
        widget=SmallTextInput(input_type="email"))



# -*- coding: utf-8 -*-
"""
Rhizom - Relationship grapher

Copyright (C) 2015  Aurelien Bompard <aurelien@bompard.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import absolute_import, unicode_literals, print_function

from flask import current_app, g
from flask.ext.login import current_user

from .models import User, Permission, PermissionLevel


def get_relationship_ids(nodeid, links):
    rel_ids = set()
    for link in links:
        if link["source"] == nodeid:
            rel_ids.add(link["target"])
        if link["target"] == nodeid:
            rel_ids.add(link["source"])
    return rel_ids

def compute_node_circles(nodes, links, center_index):
    def compute_circle(circle):
        if circle > 30:
            return # Recursion failsafe
        next_nodes = set()
        for nodeid, node in enumerate(nodes):
            if circle is not None and node.get("circle") != circle:
                continue # only consider the current circle
            next_nodes.update(get_relationship_ids(nodeid, links))
        next_nodes = [ n for n in next_nodes
                       if not nodes[n].has_key("circle") ]
        if not next_nodes:
            return # nothing left to set
        for nodeid in next_nodes:
            nodes[nodeid]["circle"] = circle + 1
        compute_circle(circle + 1)
    nodes[center_index]["circle"] = 0
    compute_circle(0)


def compute_node_branch(nodes, links):
    def compute_branch(circle):
        if circle > 30:
            return # Recursion failsafe
        next_nodes_ids = [ n["id"] for n in nodes
                           if "circle" in n and n["circle"] == circle ]
        for nodeid, node in enumerate(nodes):
            if node["id"] not in next_nodes_ids or node.has_key("branch"):
                continue
            branch = None
            for rel_id in get_relationship_ids(nodeid, links):
                rel_branch = nodes[rel_id].get("branch")
                if rel_branch is not None:
                    branch = rel_branch # TODO: allow being on multiple branches
                    #print(node["name"], "has branch", branch, "by", nodes[rel_id]["name"])
            assert branch is not None, "No branch found for %s" % node["name"]
            node["branch"] = branch
        compute_branch(circle + 1)
    first_circle = [ n for n in nodes if "circle" in n and n["circle"] == 1 ]
    for node in first_circle:
        node["branch"] = node["name"]
    compute_branch(2)


def require_permission_level(graph, level):
    if current_app.login_manager._login_disabled:
        return True
    if not current_user.is_authenticated():
        return False
    if current_user.is_master:
        return True
    perm = g.db.query(Permission).filter(
        graph_id=graph, user_email=current_user.email
        ).order_by(Permission.level.desc()).first()
    if perm is None:
        return False
    return perm.level >= level



# -*- coding: utf-8 -*-
"""
Rhizom - Relationship grapher

Copyright (C) 2015  Aurelien Bompard <aurelien@bompard.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import with_statement, unicode_literals, print_function

from alembic import context
from sqlalchemy import create_engine, pool
#from logging.config import fileConfig

# Interpret the config file for Python logging.
# This line sets up loggers basically.
#fileConfig(config.config_file_name)

try:
    import rhizom
except ImportError:
    import os, sys
    if "rhizom" not in os.listdir("."):
        raise
    sys.path.append(os.getcwd())
from rhizom import app
from rhizom.models import Base

url = app.config["SQLALCHEMY_DATABASE_URI"]
target_metadata = Base.metadata


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(url=url, target_metadata=target_metadata)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    engine = create_engine(url, poolclass=pool.NullPool)
    connection = engine.connect()
    context.configure(
        connection=connection,
        target_metadata=target_metadata
    )

    try:
        with context.begin_transaction():
            context.run_migrations()
    finally:
        connection.close()

if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# -*- coding: utf-8 -*-
"""
Rhizom - Relationship grapher

Copyright (C) 2015  Aurelien Bompard <aurelien@bompard.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import absolute_import, unicode_literals, print_function

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from alembic import command
from alembic.config import Config as AlembicConfig


def get_session(url):
    engine = create_engine(url, convert_unicode=True)
    session = scoped_session(sessionmaker(
                    autocommit=False, autoflush=False, bind=engine))
    from .models import Base
    Base.query = session.query_property()
    return session


def init_db(url):
    # import all modules here that might define models so that
    # they will be registered properly on the metadata.  Otherwise
    # you will have to import them first before calling init_db()
    engine = create_engine(url, convert_unicode=True)
    from .models import Base
    Base.metadata.create_all(bind=engine)
    alembic_cfg = AlembicConfig()
    alembic_cfg.set_main_option("script_location", "rhizom:migrations")
    command.stamp(alembic_cfg, "head")




# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import os
basedir = os.path.abspath(os.path.dirname(__file__))

DEBUG = True
SQLALCHEMY_DATABASE_URI = 'sqlite:///' + os.path.join(basedir, '..', 'rhizom.db')
BROWSERID_AUDIENCE = ["http://127.0.0.1:5000", "http://127.0.0.1:5001"]
SECRET_KEY = 'SomethingVerySecretThatYouMustChange'
ADMINS = ("you@example.com")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    import setuptools
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()

from setuptools import setup, find_packages

def reqfile(filepath):
    """Turns a text file into a list (one element per line)"""
    result = []
    import re
    url_re = re.compile(".+:.+#egg=(.+)")
    with open(filepath, "r") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            mo = url_re.match(line)
            if mo is not None:
                line = mo.group(1)
            result.append(line)
    return result


setup(
    name="rhizom",
    version="0.1",
    description="Visualize your relationships",
    long_description=open('README.md').read(),
    author='Aurelien Bompard',
    author_email='aurelien@bompard.org',
    url="https://gitlab.com/abompard/rhizom",
    license="AGPLv3+",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Programming Language :: Python :: 2",
        ],
    packages=find_packages(),
    #include_package_data=True,
    install_requires=reqfile("requirements.txt"),
    #test_suite = "rhizom.test",
    entry_points={
        'console_scripts': [
            'rhizom = rhizom.scripts:main',
            ],
        },
    )



# -*- coding: utf-8 -*-
"""
Rhizom - Relationship grapher

Copyright (C) 2015  Aurelien Bompard <aurelien@bompard.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import unicode_literals, print_function

import os
import json
from optparse import OptionParser

from sqlalchemy.orm.exc import NoResultFound
from alembic import command
from alembic.config import Config as AlembicConfig

import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from rhizom import app
from rhizom.database import init_db, get_session
from rhizom.models import Person, Relationship, User



def initdb(args):
    init_db(app.config["SQLALCHEMY_DATABASE_URI"])


def runserver(args):
    if args:
        app.run(port=int(args[0]))
    else:
        app.run()


def main():
    if len(sys.argv) < 2:
        print("wrong number of arguments.", file=sys.stderr)
        sys.exit(2)
    funcname = sys.argv[1]
    localnames = locals().keys()
    funcnames = [ n for n in localnames
                  if callable(locals()[n]) and n != "main" ]
    if funcname in ("-h", "--help"):
        print("Available scripts: {}".format(", ".join(funcnames)))
        sys.exit(2)
    if funcname not in funcnames:
        print("no such script: {}.".format(funcname), file=sys.stderr)
        sys.exit(2)
    locals()[funcname](sys.argv[2:])


if __name__ == '__main__':
    main()
